@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2013 Tim Ruehsen
- * Copyright (c) 2015-2022 Free Software Foundation, Inc.
+ * Copyright (c) 2015-2024 Free Software Foundation, Inc.
  *
  * This file is part of libwget.
  *
@@ -249,7 +249,8 @@ static void html_get_url(void *context, int flags, const char *tag, const char *
 				if (ctx->uri_index >= 0) {
 					// href= came before rel=
 					wget_html_parsed_url *url = wget_vector_get(res->uris, ctx->uri_index);
-					url->link_inline = ctx->link_inline;
+					if (url)
+						url->link_inline = ctx->link_inline;
 				}
 				return;
 			}
@@ -258,9 +259,6 @@ static void html_get_url(void *context, int flags, const char *tag, const char *
 		if ((*tag|0x20) == 'a' && (tag[1] == 0 || !wget_strcasecmp_ascii(tag, "area"))
 			&& !wget_strcasecmp_ascii(attr, "download"))
 		{
-			if (!val)
-				return;
-
 			for (;len && c_isspace(*val); val++, len--); // skip leading spaces
 			for (;len && c_isspace(val[len - 1]); len--);  // skip trailing spaces
 			if (!len)
