@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2012 Tim Ruehsen
- * Copyright (c) 2015-2021 Free Software Foundation, Inc.
+ * Copyright (c) 2015-2024 Free Software Foundation, Inc.
  *
  * This file is part of libwget.
  *
@@ -38,6 +38,8 @@
 #include <wget.h>
 #include "private.h"
 
+static char *create_safe_uri(wget_iri *iri);
+
 /**
  * \file
  * \brief Functions to work with URIs and IRIs
@@ -82,6 +84,8 @@ static struct iri_scheme {
 	[WGET_IRI_SCHEME_HTTPS] = { 443, "https" },
 };
 
+static size_t WGET_GCC_NONNULL_ALL normalize_path(char *path);
+
 /**
  * \param[in] scheme Scheme to get name for
  * \return Name of \p scheme (e.g. "http" or "https") or NULL is not supported
@@ -291,41 +295,13 @@ static char *iri_unescape_inline(char *src, int ctype)
 					continue;
 				}
 			}
-		} else if (*s == '&') {
-			// entities are case sensitive (RFC1866, 3.2.3)
-			if (!strncmp((char *) s + 1, "amp;", 4)) {
-				*d++ = '&';
-				s += 5;
-				ret = src;
-				continue;
-			} else if (!strncmp((char *) s + 1, "gt;", 3)) {
-				*d++ = '>';
-				s += 4;
-				ret = src;
-				continue;
-			} else if (!strncmp((char *) s + 1, "lt;", 3)) {
-				*d++ = '<';
-				s += 4;
-				ret = src;
-				continue;
-			} else if (!strncmp((char *) s + 1, "quot;", 5)) {
-				*d++ = '\"';
-				s += 6;
-				ret = src;
-				continue;
-			} else if (!strncmp((char *) s + 1, "apos;", 5)) {
-				*d++ = '\'';
-				s += 6;
-				ret = src;
-				continue;
-			}
 		} else if (*s == '#') {
 			uint32_t value = 0;
 
 			if (s[1] == 'x') {
 				unsigned char *p = s + 2;
 				while (c_isxdigit(*p)) {
-					value = (value << 4) | unhex(*p);
+					value = ((value & 0x0FFFFFFF) << 4) | unhex(*p);
 					p++;
 				}
 				if (*p == ';') {
@@ -416,8 +392,10 @@ char *wget_iri_unescape_url_inline(char *src)
 void wget_iri_free_content(wget_iri *iri)
 {
 	if (iri) {
-		if (iri->uri_allocated)
-			xfree(iri->uri);
+		if (iri->userinfo)
+			xfree(iri->safe_uri);
+		else
+			iri->safe_uri = NULL;
 		if (iri->host_allocated)
 			xfree(iri->host);
 		if (iri->path_allocated)
@@ -516,18 +494,19 @@ wget_iri *wget_iri_parse(const char *url, const char *encoding)
 
 	// just use one block of memory for all parsed URI parts
 	slen = strlen(url);
-	extra = have_scheme ? 0 : sizeof("http://") - 1; // extra space for http://
 
-	iri = wget_malloc(sizeof(wget_iri) + (slen + extra + 1) * 2);
+	// Calculate extra space for inserting a scheme. Take the longest supported scheme string.
+	extra = have_scheme ? 0 : sizeof("https://") - 1;
+
+	iri = wget_calloc(1, sizeof(wget_iri) + (slen + extra + 1) * 2);
 	if (!iri)
 		return NULL;
 
-	memset(iri, 0, sizeof(wget_iri));
+	iri->msize = slen + extra + 1;
 
 	if (have_scheme) {
-		iri->msize = slen + 1;
-		iri->uri = memcpy(((char *)iri) + sizeof(wget_iri), url, iri->msize);
-		p = s = memcpy((char *)iri->uri + iri->msize, url, iri->msize);
+		iri->uri = memcpy(iri + 1, url, slen + 1);
+		p = s = memcpy((char *)iri->uri + iri->msize, url, slen + 1);
 		s = strchr(s, ':'); // we know there is a :
 		*s++ = 0;
 
@@ -555,13 +534,13 @@ wget_iri *wget_iri_parse(const char *url, const char *encoding)
 		}
 	} else {
 		// add http:// scheme to url
-		iri->uri = memcpy(((char *)iri) + sizeof(wget_iri), "http://", extra);
-		memcpy(((char *)iri) + sizeof(wget_iri) + extra, url, slen + 1);
-		iri->msize = slen + 1 + extra;
-		s = memcpy((char *)iri->uri + iri->msize, "http://", extra);
-		memcpy((char *)iri->uri + iri->msize + extra, url, slen + 1);
-		s[extra - 3] = 0;
-		s += extra;
+#define HTTP_SCHEME_STRING "http://"
+#define HTTP_SCHEME_LEN (sizeof(HTTP_SCHEME_STRING) - 1)
+		iri->uri = memcpy(iri + 1, HTTP_SCHEME_STRING, HTTP_SCHEME_LEN);
+		memcpy((char *)iri->uri + HTTP_SCHEME_LEN, url, slen + 1);
+		s = memcpy((char *)iri->uri + iri->msize, iri->uri, HTTP_SCHEME_LEN + slen + 1);
+		s[HTTP_SCHEME_LEN - 3] = 0;
+		s += HTTP_SCHEME_LEN;
 
 		iri->scheme = WGET_IRI_SCHEME_HTTP;
 		iri->port = schemes[WGET_IRI_SCHEME_HTTP].port;
@@ -590,6 +569,7 @@ wget_iri *wget_iri_parse(const char *url, const char *encoding)
 		c = *s;
 		if (c) *s++ = 0;
 		wget_iri_unescape_inline((char *)iri->path);
+		normalize_path((char *)iri->path);
 	}
 
 	if (c == '?') {
@@ -644,11 +624,14 @@ wget_iri *wget_iri_parse(const char *url, const char *encoding)
 		}
 		if (*s == ':') {
 			if (c_isdigit(s[1])) {
-				int port = atoi(s + 1);
-				if (port > 0 && port < 65536) {
-					iri->port = (uint16_t) port;
-					iri->port_given = true;
+				unsigned long port = strtoul(s + 1, NULL, 10);
+				if (port == 0 || port > 65535) {
+					error_printf(_("Port number must be in the range 1..65535\n"));
+					wget_iri_free(&iri);
+					return NULL;
 				}
+				iri->port = (uint16_t) port;
+				iri->port_given = true;
 			}
 		}
 		*s = 0;
@@ -703,6 +686,12 @@ wget_iri *wget_iri_parse(const char *url, const char *encoding)
 		}
 	}
 
+	if (iri->userinfo) {
+		iri->safe_uri = create_safe_uri(iri);
+	} else {
+		iri->safe_uri = iri->uri;
+	}
+
 /*
 	debug_printf("scheme=%s\n",iri->scheme);
 	debug_printf("host=%s\n",iri->host);
@@ -710,7 +699,6 @@ wget_iri *wget_iri_parse(const char *url, const char *encoding)
 	debug_printf("query=%s\n",iri->query);
 	debug_printf("fragment=%s\n",iri->fragment);
 */
-
 	return iri;
 }
 
@@ -725,16 +713,17 @@ wget_iri *wget_iri_clone(const wget_iri *iri)
 	if (!iri || !iri->uri)
 		return NULL;
 
-	size_t slen = strlen(iri->uri);
-	wget_iri *clone = wget_malloc(sizeof(wget_iri) + (slen + 1) + iri->msize);
+	wget_iri *clone = wget_memdup(iri, sizeof(wget_iri) + iri->msize * 2);
 
 	if (!clone)
 		return NULL;
 
-	memcpy(clone, iri, sizeof(wget_iri));
-	clone->uri = memcpy(((char *)clone) + sizeof(wget_iri), iri->uri, slen + 1);
-	memcpy((char *)clone->uri + slen + 1, (char *)iri->uri + slen + 1, iri->msize);
-	clone->uri_allocated = 0;
+	clone->uri = (char *) (clone + 1);
+
+	if (iri->userinfo)
+		clone->safe_uri = wget_strdup(iri->safe_uri);
+	else
+		clone->safe_uri = clone->uri;
 
 	clone->connection_part = wget_strdup(iri->connection_part);
 
@@ -785,11 +774,13 @@ wget_iri *wget_iri_clone(const wget_iri *iri)
 const char *wget_iri_get_connection_part(const wget_iri *iri, wget_buffer *buf)
 {
 	if (iri) {
-		if (iri->port_given) {
-			wget_buffer_printf_append(buf, "%s://%s:%hu", schemes[iri->scheme].name, iri->host, iri->port);
-		} else {
+		if (wget_ip_is_family(iri->host, WGET_NET_FAMILY_IPV6))
+			wget_buffer_printf_append(buf, "%s://[%s]", schemes[iri->scheme].name, iri->host);
+		else
 			wget_buffer_printf_append(buf, "%s://%s", schemes[iri->scheme].name, iri->host);
-		}
+
+		if (iri->port_given)
+			wget_buffer_printf_append(buf, ":%hu", iri->port);
 	}
 
 	return buf->data;
@@ -897,14 +888,18 @@ static size_t WGET_GCC_NONNULL_ALL normalize_path(char *path)
  */
 const char *wget_iri_relative_to_abs(const wget_iri *base, const char *val, size_t len, wget_buffer *buf)
 {
-	debug_printf("*url = %.*s\n", (int)len, val);
-
 	if (len == (size_t) -1)
 		len = strlen(val);
 
 	if (*val == '/') {
 		if (base) {
-			char path[len + 1];
+			char tmp[4096], *path = tmp;
+
+			if (len >= sizeof(tmp)) {
+				path = wget_malloc(len + 1);
+				if (!path)
+					return NULL;
+			}
 
 			// strlcpy or snprintf are ineffective here since they do strlen(val), which might be large
 			wget_strscpy(path, val, len + 1);
@@ -919,7 +914,6 @@ const char *wget_iri_relative_to_abs(const wget_iri *base, const char *val, size
 				wget_buffer_strcpy(buf, schemes[base->scheme].name);
 				wget_buffer_strcat(buf, ":");
 				wget_buffer_strcat(buf, path);
-				debug_printf("*1 %s\n", buf->data);
 			} else {
 				// absolute path
 				normalize_path(path);
@@ -928,8 +922,10 @@ const char *wget_iri_relative_to_abs(const wget_iri *base, const char *val, size
 				wget_iri_get_connection_part(base, buf);
 				wget_buffer_strcat(buf, "/");
 				wget_buffer_strcat(buf, path);
-				debug_printf("*2 %s\n", buf->data);
 			}
+
+			if (path != tmp)
+				xfree(path);
 		} else {
 			return NULL;
 		}
@@ -939,9 +935,7 @@ const char *wget_iri_relative_to_abs(const wget_iri *base, const char *val, size
 			// absolute URI
 			if (buf) {
 				wget_buffer_memcpy(buf, val, len);
-				debug_printf("*3 %s\n", buf->data);
 			} else {
-				debug_printf("*3 %s\n", val);
 				return val;
 			}
 		} else if (base) {
@@ -960,8 +954,6 @@ const char *wget_iri_relative_to_abs(const wget_iri *base, const char *val, size
 				wget_buffer_memcat(buf, val, len);
 
 			buf->length = normalize_path(buf->data + tmp_len) + tmp_len;
-
-			debug_printf("*4 %s %zu\n", buf->data, buf->length);
 		} else if (val[len] == 0) {
 			return val;
 		} else {
@@ -1021,7 +1013,7 @@ wget_iri *wget_iri_parse_base(const wget_iri *base, const char *url, const char
  * This function uses wget_strcasecmp() to compare the various parts of the IRIs so a non-zero negative return value
  * indicates that \p iri1 is less than \p iri2, whereas a positive value indicates \p iri1 is greater than \p iri2.
  */
-int wget_iri_compare(wget_iri *iri1, wget_iri *iri2)
+int wget_iri_compare(const wget_iri *iri1, const wget_iri *iri2)
 {
 	int n;
 
@@ -1210,10 +1202,7 @@ const char *wget_iri_get_escaped_resource(const wget_iri *iri, wget_buffer *buf)
 	if (iri->query) {
 		wget_buffer_memcat(buf, "?", 1);
 		for (const char *p = iri->query; *p; p++)
-			if (*p == ' ')
-				wget_buffer_memcat(buf, "%20", 3);
-			else
-				wget_buffer_memcat(buf, p, 1);
+			wget_buffer_memcat(buf, *p == ' ' ? "+" : p, 1);
 	}
 
 	return buf->data;
@@ -1452,17 +1441,53 @@ wget_iri_scheme wget_iri_set_scheme(wget_iri *iri, wget_iri_scheme scheme)
 			iri->port = schemes[scheme].port;
 
 		size_t old_scheme_len = strlen(schemes[old_scheme].name);
+		size_t scheme_len = strlen(schemes[scheme].name);
 
 		if (strncmp(iri->uri, schemes[old_scheme].name, old_scheme_len) == 0 && iri->uri[old_scheme_len] == ':') {
-			char *new_uri = wget_aprintf("%s%s",  schemes[iri->scheme].name, iri->uri + old_scheme_len);
-			if (iri->uri_allocated)
-				xfree(iri->uri);
-			iri->uri = new_uri;
-			iri->uri_allocated = true;
+			memmove((char *)iri->uri + scheme_len, iri->uri + old_scheme_len, iri->msize - old_scheme_len);
+			memcpy((char *)iri->uri, schemes[scheme].name, scheme_len);
+		}
+
+		if (iri->userinfo) {
+			xfree(iri->safe_uri);
+			iri->safe_uri = create_safe_uri(iri);
+		} else {
+			iri->safe_uri = iri->uri;
 		}
 	}
 
 	return old_scheme;
 }
 
+static char *create_safe_uri(wget_iri *iri)
+{
+	if (!iri || !iri->uri)
+		return NULL;
+
+	wget_buffer *buf = wget_buffer_alloc(strlen(iri->uri));
+	if (!buf)
+		return NULL;
+
+	wget_buffer_printf(buf, "%s://%s", schemes[iri->scheme].name, iri->host);
+
+	if (iri->path) {
+		wget_buffer_strcat(buf, "/");
+		wget_buffer_strcat(buf, iri->path);
+	}
+	if (iri->query) {
+		wget_buffer_strcat(buf, "?");
+		wget_buffer_strcat(buf, iri->query);
+	}
+	if (iri->fragment) {
+		wget_buffer_strcat(buf, "#");
+		wget_buffer_strcat(buf, iri->fragment);
+	}
+
+	char *safe_uri = buf->data;
+	buf->data = NULL;
+	wget_buffer_free(&buf);
+
+	return safe_uri;
+}
+
 /** @} */
