@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2013-2014 Tim Ruehsen
- * Copyright (c) 2015-2023 Free Software Foundation, Inc.
+ * Copyright (c) 2015-2024 Free Software Foundation, Inc.
  *
  * This file is part of Wget
  *
@@ -37,6 +37,7 @@
 #include <fcntl.h>
 #include <utime.h>
 #include <dirent.h>
+#include <ctype.h>
 #include <sys/stat.h>
 #include <sys/wait.h>
 
@@ -45,12 +46,14 @@
 #include "libtest.h"
 
 #include <microhttpd.h>
-#ifdef HAVE_MICROHTTPD_HTTP2_H
-#  include <microhttpd_http2.h>
-#endif
 #ifndef HAVE_MHD_FREE
 #  define MHD_free wget_free
 #endif
+
+#ifdef WITH_LIBNGHTTP2
+#  include <nghttp2/nghttp2.h>
+#  include <pthread.h>
+#endif
 #ifndef MHD_HTTP_RANGE_NOT_SATISFIABLE
 #  define MHD_HTTP_RANGE_NOT_SATISFIABLE MHD_HTTP_REQUESTED_RANGE_NOT_SATISFIABLE
 #endif
@@ -67,7 +70,6 @@ enum MHD_Result {
 #endif
 
 #include <sys/types.h>
-#include <sys/select.h>
 #include <sys/socket.h>
 #include <netdb.h>
 
@@ -90,9 +92,12 @@ static int
 	keep_tmpfiles,
 	clean_directory,
 	reject_http_connection,
-	reject_https_connection;
+	reject_https_connection,
+	ocsp_response_pos,
+	h2_min_concurrent_streams;
 static wget_vector
-	*request_urls;
+	*request_urls,
+	*ocsp_responses;
 static wget_test_url_t
 	*urls;
 static size_t
@@ -114,19 +119,18 @@ static enum CHECK_POST_HANDSHAKE_AUTH {
 static struct MHD_Daemon
 	*httpdaemon,
 	*httpsdaemon,
-	*ocspdaemon,
-	*h2daemon;
+	*ocspdaemon;
 
 #ifdef WITH_GNUTLS_OCSP
 static gnutls_pcert_st *pcrt;
 static gnutls_privkey_t *privkey;
 
-static struct ocsp_resp_t {
+typedef struct {
 	char
 		*data;
 	size_t
 		size;
-} *ocsp_resp;
+} ocsp_resp_t;
 #endif
 
 #ifdef WITH_GNUTLS_OCSP
@@ -135,6 +139,32 @@ static gnutls_ocsp_data_st *ocsp_stap_resp;
 #endif
 #endif
 
+#ifdef WITH_LIBNGHTTP2
+typedef struct {
+	nghttp2_session *session;
+	gnutls_session_t tls_session;
+	int sockfd;
+	char *method;
+	char *path;
+	const char *response_body;  // Points to test URL body
+	size_t response_body_len;
+	size_t response_body_offset;
+	int32_t stream_id;
+	char **header_names;   // Allocated header names to free later
+	char **header_values;  // Allocated header values to free later
+	int num_headers;       // Number of allocated headers
+} h2_session_data;
+
+typedef struct {
+	int listen_fd;
+	pthread_t thread;
+	volatile int running;
+	gnutls_certificate_credentials_t creds;
+} h2_server_t;
+
+static h2_server_t *nghttp2_server;
+#endif
+
 // for passing URL query string
 struct query_string {
 	wget_buffer
@@ -243,7 +273,7 @@ static ssize_t _callback (void *cls, uint64_t pos, char *buf, size_t buf_size)
 		(struct ResponseContentCallbackParam *)cls;
 
 	if (pos >= param->response_size)
-		return (ssize_t) MHD_CONTENT_READER_END_OF_STREAM;
+		return MHD_CONTENT_READER_END_OF_STREAM;
 
 	// divide data into two chunks
 	buf_size = (param->response_size / 2) + 1;
@@ -264,7 +294,7 @@ static ssize_t _callback_interruptable (void *cls, uint64_t pos, char *buf, size
 		(struct ResponseContentCallbackParam *)cls;
 
 	if (pos >= param->response_size)
-		return (ssize_t) MHD_CONTENT_READER_END_OF_STREAM;
+		return MHD_CONTENT_READER_END_OF_STREAM;
 
 	if (buf_size <= (param->response_size - pos)) {
 		size_to_copy = buf_size;
@@ -274,7 +304,7 @@ static ssize_t _callback_interruptable (void *cls, uint64_t pos, char *buf, size
 
 	if (param->interrupt_response_mode != INTERRUPT_RESPONSE_DISABLED) {
 		if (pos >= param->interrupt_response_after_nbytes) {
-			return (ssize_t) MHD_CONTENT_READER_END_WITH_ERROR;
+			return MHD_CONTENT_READER_END_WITH_ERROR;
 		}
 
 		if (size_to_copy > (param->interrupt_response_after_nbytes - pos)) {
@@ -311,14 +341,14 @@ static enum MHD_Result _ocsp_ahc(
 	} else if (!first && upload_data == NULL) {
 		int ret = 0;
 
-		if (ocsp_resp->data) {
+		ocsp_resp_t *ocsp_resp = wget_vector_get(ocsp_responses, ocsp_response_pos++);
+
+		if (ocsp_resp) {
 			struct MHD_Response *response = MHD_create_response_from_buffer (ocsp_resp->size, ocsp_resp->data, MHD_RESPMEM_MUST_COPY);
 
 			ret = MHD_queue_response (connection, MHD_HTTP_OK, response);
 
 			MHD_destroy_response (response);
-
-			wget_xfree(ocsp_resp->data);
 		}
 
 		return ret;
@@ -414,7 +444,7 @@ static enum MHD_Result _answer_to_connection(
 	// get query string
 	query.params = wget_buffer_alloc(1024);
 	query.it = 0;
-	MHD_get_connection_values(connection, MHD_GET_ARGUMENT_KIND, (MHD_KeyValueIterator)_print_query_string, &query);
+	MHD_get_connection_values(connection, MHD_GET_ARGUMENT_KIND, _print_query_string, &query);
 
 	// get if-modified-since header
 	modified_val = MHD_lookup_connection_value(connection, MHD_HEADER_KIND,
@@ -426,7 +456,7 @@ static enum MHD_Result _answer_to_connection(
 	// get header range
 	wget_buffer *header_range = wget_buffer_alloc(1024);
 	if (!strcmp(method, "GET"))
-		MHD_get_connection_values(connection, MHD_HEADER_KIND, (MHD_KeyValueIterator)_print_header_range, header_range);
+		MHD_get_connection_values(connection, MHD_HEADER_KIND, _print_header_range, header_range);
 
 	from_bytes = to_bytes = 0;
 	if (*header_range->data) {
@@ -703,24 +733,497 @@ static enum MHD_Result _answer_to_connection(
 	return ret;
 }
 
+#ifdef WITH_LIBNGHTTP2
+
+static ssize_t h2_send_callback(nghttp2_session *session WGET_GCC_UNUSED,
+	const uint8_t *data, size_t length, int flags WGET_GCC_UNUSED, void *user_data)
+{
+	h2_session_data *session_data = (h2_session_data *) user_data;
+	ssize_t nwrite = gnutls_record_send(session_data->tls_session, data, length);
+
+	if (nwrite < 0) {
+		if (nwrite == GNUTLS_E_AGAIN || nwrite == GNUTLS_E_INTERRUPTED)
+			return NGHTTP2_ERR_WOULDBLOCK;
+		return NGHTTP2_ERR_CALLBACK_FAILURE;
+	}
+
+	return nwrite;
+}
+
+static ssize_t h2_recv_callback(nghttp2_session *session WGET_GCC_UNUSED,
+	uint8_t *buf, size_t length, int flags WGET_GCC_UNUSED, void *user_data)
+{
+	h2_session_data *session_data = (h2_session_data *) user_data;
+	ssize_t nread = gnutls_record_recv(session_data->tls_session, buf, length);
+
+	if (nread < 0) {
+		if (nread == GNUTLS_E_AGAIN || nread == GNUTLS_E_INTERRUPTED)
+			return NGHTTP2_ERR_WOULDBLOCK;
+		return NGHTTP2_ERR_CALLBACK_FAILURE;
+	}
+
+	if (nread == 0)
+		return NGHTTP2_ERR_EOF;
+
+	return nread;
+}
+
+static ssize_t h2_data_read_callback(nghttp2_session *session,
+	int32_t stream_id, uint8_t *buf, size_t length, uint32_t *data_flags,
+	nghttp2_data_source *source WGET_GCC_UNUSED, void *user_data WGET_GCC_UNUSED)
+{
+	h2_session_data *stream_data = nghttp2_session_get_stream_user_data(session, stream_id);
+	if (!stream_data || !stream_data->response_body)
+		return 0;
+
+	size_t remaining = stream_data->response_body_len - stream_data->response_body_offset;
+
+	// Limit chunk size to 8KB to allow proper rate limiting on the client side
+	// Without this limit, large responses are sent too quickly and bypass rate limiting
+	// Smaller chunks allow better rate limiting granularity, especially with HTTP/2 multiplexing
+	static const size_t max_chunk = 8 * 1024;
+	if (length > max_chunk)
+		length = max_chunk;
+
+	size_t to_send = remaining < length ? remaining : length;
+
+	if (to_send > 0) {
+		memcpy(buf, stream_data->response_body + stream_data->response_body_offset, to_send);
+		stream_data->response_body_offset += to_send;
+	}
+
+	if (stream_data->response_body_offset >= stream_data->response_body_len)
+		*data_flags |= NGHTTP2_DATA_FLAG_EOF;
+
+	return (ssize_t) to_send;
+}
+
+static int h2_on_header_callback(nghttp2_session *session,
+	const nghttp2_frame *frame, const uint8_t *name, size_t namelen,
+	const uint8_t *value, size_t valuelen, uint8_t flags WGET_GCC_UNUSED, void *user_data WGET_GCC_UNUSED)
+{
+	h2_session_data *session_data = nghttp2_session_get_stream_user_data(session, frame->hd.stream_id);
+
+	if (!session_data || frame->hd.type != NGHTTP2_HEADERS)
+		return 0;
+
+	if (frame->headers.cat != NGHTTP2_HCAT_REQUEST)
+		return 0;
+
+	// Parse :method and :path pseudo-headers
+	if (namelen == 7 && memcmp(name, ":method", 7) == 0) {
+		session_data->method = wget_strmemdup((const char *) value, valuelen);
+	} else if (namelen == 5 && memcmp(name, ":path", 5) == 0) {
+		session_data->path = wget_strmemdup((const char *) value, valuelen);
+	}
+
+	return 0;
+}
+
+static int h2_on_begin_headers_callback(nghttp2_session *session,
+	const nghttp2_frame *frame, void *user_data)
+{
+	if (frame->hd.type != NGHTTP2_HEADERS || frame->headers.cat != NGHTTP2_HCAT_REQUEST)
+		return 0;
+
+	h2_session_data *stream_data = wget_calloc(1, sizeof(h2_session_data));
+	if (!stream_data)
+		return NGHTTP2_ERR_CALLBACK_FAILURE;
+
+	h2_session_data *session_data = (h2_session_data *)user_data;
+	stream_data->session = session;
+	stream_data->tls_session = session_data->tls_session;
+	stream_data->sockfd = session_data->sockfd;
+	stream_data->stream_id = frame->hd.stream_id;
+	stream_data->response_body = NULL;
+	stream_data->response_body_len = 0;
+	stream_data->response_body_offset = 0;
+
+	nghttp2_session_set_stream_user_data(session, frame->hd.stream_id, stream_data);
+
+	return 0;
+}
+
+static int h2_on_frame_recv_callback(nghttp2_session *session,
+	const nghttp2_frame *frame, void *user_data WGET_GCC_UNUSED)
+{
+	if (frame->hd.type != NGHTTP2_HEADERS || frame->headers.cat != NGHTTP2_HCAT_REQUEST)
+		return 0;
+
+	h2_session_data *stream_data = nghttp2_session_get_stream_user_data(session, frame->hd.stream_id);
+	if (!stream_data || !stream_data->path)
+		return 0;
+
+	// Find matching URL in test definitions
+	const char *path = stream_data->path;
+	if (*path == '/')
+		path++;
+
+	wget_test_url_t *url = NULL;
+	for (size_t i = 0; i < nurls; i++) {
+		const char *url_name = urls[i].name;
+		if (*url_name == '/')
+			url_name++;
+
+		if (strcmp(path, url_name) == 0) {
+			url = &urls[i];
+			break;
+		}
+	}
+
+	// Prepare response
+	const char *status = "404 Not Found";
+	const char *body = "Not Found";
+	size_t body_len = strlen(body);
+
+	if (url) {
+		status = url->code;
+		body = url->body;
+		body_len = url->body_len ? url->body_len : (body ? strlen(body) : 0);
+	}
+
+	// Store response body in stream context
+	stream_data->response_body = body;
+	stream_data->response_body_len = body_len;
+	stream_data->response_body_offset = 0;
+
+	// Parse status code
+	int status_code = 404;
+	if (status)
+		status_code = atoi(status);
+
+	// Build response headers
+	char status_str[16];
+	wget_snprintf(status_str, sizeof(status_str), "%d", status_code);
+
+	#define MAX_HEADERS 20
+	nghttp2_nv hdrs[MAX_HEADERS];
+	int nhdrs = 0;
+
+	hdrs[nhdrs].name = (uint8_t *) ":status";
+	hdrs[nhdrs].namelen = strlen((char *) hdrs[nhdrs].name);
+	hdrs[nhdrs].value = (uint8_t *) status_str;
+	hdrs[nhdrs].valuelen = strlen(status_str);
+	hdrs[nhdrs].flags = NGHTTP2_NV_FLAG_NONE;
+	nhdrs++;
+
+	hdrs[nhdrs].name = (uint8_t *) "server";
+	hdrs[nhdrs].namelen = strlen((char *) hdrs[nhdrs].name);
+	hdrs[nhdrs].value = (uint8_t *) "libwget-test-h2/1.0";
+	hdrs[nhdrs].valuelen = strlen((char *) hdrs[nhdrs].value);
+	hdrs[nhdrs].flags = NGHTTP2_NV_FLAG_NONE;
+	nhdrs++;
+
+	// Add custom headers from test definition
+	if (url) {
+		for (int i = 0; i < 10 && url->headers[i] && nhdrs < MAX_HEADERS - 1; i++) {
+			const char *header = url->headers[i];
+			const char *colon = strchr(header, ':');
+			if (colon) {
+				size_t name_len = colon - header;
+				const char *value = colon + 1;
+				while (*value == ' ') value++;
+
+				// Convert header name to lowercase for HTTP/2
+				char *name_lower = wget_strmemdup(header, name_len);
+				for (size_t j = 0; j < name_len; j++)
+					name_lower[j] = (char) tolower((unsigned char) name_lower[j]);
+
+				hdrs[nhdrs].name = (uint8_t *) name_lower;
+				hdrs[nhdrs].namelen = name_len;
+				hdrs[nhdrs].value = (uint8_t *) wget_strdup(value);
+				hdrs[nhdrs].valuelen = strlen(value);
+				hdrs[nhdrs].flags = NGHTTP2_NV_FLAG_NONE;
+				nhdrs++;
+			}
+		}
+	}
+
+	// Setup data provider for response body
+	nghttp2_data_provider data_prov;
+	data_prov.read_callback = h2_data_read_callback;
+
+	// Store allocated headers in stream_data so we can free them later
+	// (nghttp2 doesn't copy header data, so they must remain valid)
+	if (nhdrs > 2) {
+		stream_data->num_headers = nhdrs - 2;  // Exclude :status and server headers
+		stream_data->header_names = wget_malloc(stream_data->num_headers * sizeof(char *));
+		stream_data->header_values = wget_malloc(stream_data->num_headers * sizeof(char *));
+		for (int i = 0; i < stream_data->num_headers; i++) {
+			stream_data->header_names[i] = (char *) hdrs[i + 2].name;
+			stream_data->header_values[i] = (char *) hdrs[i + 2].value;
+		}
+	} else {
+		stream_data->header_names = NULL;
+		stream_data->header_values = NULL;
+		stream_data->num_headers = 0;
+	}
+
+	// Submit response with data provider
+	int rv = nghttp2_submit_response(session, stream_data->stream_id, hdrs, nhdrs,
+		body_len > 0 ? &data_prov : NULL);
+
+	return (rv == 0) ? 0 : NGHTTP2_ERR_CALLBACK_FAILURE;
+}
+
+static int h2_on_stream_close_callback(nghttp2_session *session,
+	int32_t stream_id, uint32_t error_code WGET_GCC_UNUSED, void *user_data WGET_GCC_UNUSED)
+{
+	h2_session_data *stream_data = nghttp2_session_get_stream_user_data(session, stream_id);
+
+	if (stream_data) {
+		wget_free(stream_data->method);
+		wget_free(stream_data->path);
+
+		// Free allocated header strings
+		if (stream_data->header_names) {
+			for (int i = 0; i < stream_data->num_headers; i++) {
+				wget_free(stream_data->header_names[i]);
+				wget_free(stream_data->header_values[i]);
+			}
+			wget_free(stream_data->header_names);
+			wget_free(stream_data->header_values);
+		}
+
+		wget_free(stream_data);
+	}
+
+	return 0;
+}
+
+static void *h2_server_thread(void *arg WGET_GCC_UNUSED)
+{
+	if (!nghttp2_server)
+		return NULL;
+
+	while (nghttp2_server->running) {
+		struct sockaddr_in client_addr;
+		socklen_t client_len = sizeof(client_addr);
+
+                if (wget_ready_2_read(nghttp2_server->listen_fd, 100) <= 0)
+                    continue;
+
+		int client_fd = accept(nghttp2_server->listen_fd, (struct sockaddr *) &client_addr, &client_len);
+		if (client_fd < 0)
+			continue;
+
+		// Setup TLS (keep blocking for handshake)
+		gnutls_session_t tls_session;
+		gnutls_init(&tls_session, GNUTLS_SERVER);
+		gnutls_priority_set_direct(tls_session, "NORMAL:-VERS-ALL:+VERS-TLS1.3:+VERS-TLS1.2", NULL);
+		gnutls_credentials_set(tls_session, GNUTLS_CRD_CERTIFICATE, nghttp2_server->creds);
+
+		// ALPN for h2
+		gnutls_datum_t alpn_protos[1];
+		alpn_protos[0].data = (unsigned char *)"h2";
+		alpn_protos[0].size = 2;
+		gnutls_alpn_set_protocols(tls_session, alpn_protos, 1, 0);
+
+		gnutls_transport_set_int(tls_session, client_fd);
+		gnutls_handshake_set_timeout(tls_session, GNUTLS_DEFAULT_HANDSHAKE_TIMEOUT);
+
+		// TLS handshake (blocking)
+		int ret = gnutls_handshake(tls_session);
+		if (ret < 0) {
+			gnutls_deinit(tls_session);
+			close(client_fd);
+			continue;
+		}
+
+		// Make socket non-blocking after handshake
+		int flags = fcntl(client_fd, F_GETFL, 0);
+		fcntl(client_fd, F_SETFL, flags | O_NONBLOCK);
+		gnutls_transport_set_int(tls_session, client_fd);
+
+		// Verify ALPN selected h2
+		gnutls_datum_t alpn;
+		ret = gnutls_alpn_get_selected_protocol(tls_session, &alpn);
+		if (ret < 0 || alpn.size != 2 || memcmp(alpn.data, "h2", 2) != 0) {
+			gnutls_bye(tls_session, GNUTLS_SHUT_RDWR);
+			gnutls_deinit(tls_session);
+			close(client_fd);
+			continue;
+		}
+
+		// Create nghttp2 session
+		h2_session_data session_data = {
+			.session = NULL,
+			.tls_session = tls_session,
+			.sockfd = client_fd,
+			.method = NULL,
+			.path = NULL,
+			.response_body = NULL,
+			.response_body_len = 0,
+			.response_body_offset = 0,
+			.stream_id = 0
+		};
+
+		nghttp2_session_callbacks *callbacks;
+		nghttp2_session_callbacks_new(&callbacks);
+		nghttp2_session_callbacks_set_send_callback(callbacks, h2_send_callback);
+		nghttp2_session_callbacks_set_on_frame_recv_callback(callbacks, h2_on_frame_recv_callback);
+		nghttp2_session_callbacks_set_on_stream_close_callback(callbacks, h2_on_stream_close_callback);
+		nghttp2_session_callbacks_set_on_header_callback(callbacks, h2_on_header_callback);
+		nghttp2_session_callbacks_set_on_begin_headers_callback(callbacks, h2_on_begin_headers_callback);
+		nghttp2_session_callbacks_set_recv_callback(callbacks, h2_recv_callback);
+
+		nghttp2_session_server_new(&session_data.session, callbacks, &session_data);
+		nghttp2_session_callbacks_del(callbacks);
+
+		// Send server connection preface (SETTINGS frame)
+		nghttp2_settings_entry iv[2] = {
+			{NGHTTP2_SETTINGS_MAX_CONCURRENT_STREAMS, 100},
+			{NGHTTP2_SETTINGS_ENABLE_PUSH, 0}
+		};
+		nghttp2_submit_settings(session_data.session, NGHTTP2_FLAG_NONE, iv, 2);
+
+		// Event loop for this connection
+		while (nghttp2_server->running) {
+			// Check if we want to read or write
+			int mode = nghttp2_session_want_read(session_data.session) != 0 ? WGET_IO_READABLE : 0;
+			mode |= nghttp2_session_want_write(session_data.session) != 0 ? WGET_IO_WRITABLE : 0;
+
+			if (mode == 0)
+				break;
+
+			int rc;
+			if ((rc = wget_ready_2_transfer(client_fd, 1000, WGET_IO_READABLE | WGET_IO_WRITABLE)) <= 0)
+				continue;
+
+			if (rc & WGET_IO_READABLE) {
+				int rv = nghttp2_session_recv(session_data.session);
+				if (rv != 0 && rv != NGHTTP2_ERR_WOULDBLOCK)
+					break;
+			}
+
+			if (rc & WGET_IO_WRITABLE) {
+				int rv = nghttp2_session_send(session_data.session);
+				if (rv != 0 && rv != NGHTTP2_ERR_WOULDBLOCK)
+					break;
+			}
+		}
+
+		// Cleanup
+		nghttp2_session_del(session_data.session);
+		gnutls_bye(tls_session, GNUTLS_SHUT_RDWR);
+		gnutls_deinit(tls_session);
+		close(client_fd);
+	}
+
+	return NULL;
+}
+
+static int h2_server_start(void)
+{
+	if (nghttp2_server)
+		return 0;  // Already running
+
+	nghttp2_server = wget_calloc(1, sizeof(h2_server_t));
+	if (!nghttp2_server)
+		return -1;
+
+	// Create listen socket
+	nghttp2_server->listen_fd = socket(AF_INET, SOCK_STREAM, 0);
+	if (nghttp2_server->listen_fd < 0) {
+		wget_free(nghttp2_server);
+		nghttp2_server = NULL;
+		return -1;
+	}
+
+	int optval = 1;
+	setsockopt(nghttp2_server->listen_fd, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval));
+
+	struct sockaddr_in server_addr;
+	memset(&server_addr, 0, sizeof(server_addr));
+	server_addr.sin_family = AF_INET;
+	server_addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
+	server_addr.sin_port = 0;  // Let OS assign port
+
+	if (bind(nghttp2_server->listen_fd, (struct sockaddr *) &server_addr, sizeof(server_addr)) < 0) {
+		close(nghttp2_server->listen_fd);
+		wget_free(nghttp2_server);
+		nghttp2_server = NULL;
+		return -1;
+	}
+
+	// Get assigned port
+	socklen_t addr_len = sizeof(server_addr);
+	if (getsockname(nghttp2_server->listen_fd, (struct sockaddr *) &server_addr, &addr_len) < 0) {
+		close(nghttp2_server->listen_fd);
+		wget_free(nghttp2_server);
+		nghttp2_server = NULL;
+		return -1;
+	}
+
+	h2_server_port = ntohs(server_addr.sin_port);
+
+	if (listen(nghttp2_server->listen_fd, 5) < 0) {
+		close(nghttp2_server->listen_fd);
+		wget_free(nghttp2_server);
+		nghttp2_server = NULL;
+		return -1;
+	}
+
+	// Setup TLS credentials
+	gnutls_certificate_allocate_credentials(&nghttp2_server->creds);
+	gnutls_certificate_set_x509_key_file(nghttp2_server->creds,
+		SRCDIR "/certs/x509-server-cert.pem",
+		SRCDIR "/certs/x509-server-key.pem",
+		GNUTLS_X509_FMT_PEM);
+
+	// Start server thread
+	nghttp2_server->running = 1;
+	if (pthread_create(&nghttp2_server->thread, NULL, h2_server_thread, NULL) != 0) {
+		gnutls_certificate_free_credentials(nghttp2_server->creds);
+		close(nghttp2_server->listen_fd);
+		wget_free(nghttp2_server);
+		nghttp2_server = NULL;
+		return -1;
+	}
+
+	wget_info_printf("HTTP/2 server started on port %d\n", h2_server_port);
+
+	return 0;
+}
+
+static void h2_server_stop(void)
+{
+	if (!nghttp2_server)
+		return;
+
+	nghttp2_server->running = 0;
+
+	// Shutdown the listen socket to unblock accept()
+	shutdown(nghttp2_server->listen_fd, SHUT_RDWR);
+
+	pthread_join(nghttp2_server->thread, NULL);
+
+	gnutls_certificate_free_credentials(nghttp2_server->creds);
+	close(nghttp2_server->listen_fd);
+
+	wget_free(nghttp2_server);
+	nghttp2_server = NULL;
+
+	h2_server_port = 0;
+}
+
+#endif // WITH_LIBNGHTTP2
+
 static void _http_server_stop(void)
 {
 	MHD_stop_daemon(httpdaemon);
 	MHD_stop_daemon(httpsdaemon);
 	MHD_stop_daemon(ocspdaemon);
-	MHD_stop_daemon(h2daemon);
+
+#ifdef WITH_LIBNGHTTP2
+	h2_server_stop();
+#endif
 
 	wget_xfree(key_pem);
 	wget_xfree(cert_pem);
 
-#ifdef WITH_GNUTLS_OCSP
-	gnutls_global_deinit();
-
-	if(ocsp_resp)
-		wget_free(ocsp_resp->data);
-
-	wget_xfree(ocsp_resp);
-#endif
+	// Don't call gnutls_global_deinit() here as it breaks subsequent tests
+	// that need TLS. GnuTLS will clean up automatically at process exit.
 }
 
 static enum MHD_Result _check_to_accept(
@@ -744,8 +1247,8 @@ static int _http_server_start(int SERVER_MODE)
 		static char rnd[8] = "realrnd"; // fixed 'random' value
 
 		httpdaemon = MHD_start_daemon(MHD_USE_SELECT_INTERNALLY,
-			port_num, (MHD_AcceptPolicyCallback)_check_to_accept,
-			(void *) (ptrdiff_t) SERVER_MODE, (MHD_AccessHandlerCallback)_answer_to_connection, NULL,
+			port_num, _check_to_accept,
+			(void *) (ptrdiff_t) SERVER_MODE, _answer_to_connection, NULL,
 			MHD_OPTION_DIGEST_AUTH_RANDOM, sizeof(rnd), rnd,
 			MHD_OPTION_NONCE_NC_SIZE, 300,
 #if MHD_VERSION >= 0x00095400
@@ -777,8 +1280,8 @@ static int _http_server_start(int SERVER_MODE)
 						| MHD_USE_POST_HANDSHAKE_AUTH_SUPPORT
 #endif
 					,
-					port_num, (MHD_AcceptPolicyCallback)_check_to_accept,
-					(void *) (ptrdiff_t) SERVER_MODE, (MHD_AccessHandlerCallback)_answer_to_connection, NULL,
+					port_num, _check_to_accept,
+					(void *) (ptrdiff_t) SERVER_MODE, _answer_to_connection, NULL,
 					MHD_OPTION_HTTPS_MEM_KEY, key_pem,
 					MHD_OPTION_HTTPS_MEM_CERT, cert_pem,
 #if MHD_VERSION >= 0x00095400
@@ -795,34 +1298,6 @@ static int _http_server_start(int SERVER_MODE)
 					return 1;
 				}
 			}
-			else {
-#ifdef HAVE_MICROHTTPD_HTTP2_H
-				h2daemon = MHD_start_daemon(MHD_USE_HTTP2 | MHD_USE_SELECT_INTERNALLY | MHD_USE_TLS
-#if MHD_VERSION >= 0x00096302
-						| MHD_USE_POST_HANDSHAKE_AUTH_SUPPORT
-#endif
-					,
-					port_num, (MHD_AcceptPolicyCallback)_check_to_accept,
-					(void *) (ptrdiff_t) SERVER_MODE, (MHD_AccessHandlerCallback)_answer_to_connection, NULL,
-					MHD_OPTION_HTTPS_MEM_KEY, key_pem,
-					MHD_OPTION_HTTPS_MEM_CERT, cert_pem,
-#if MHD_VERSION >= 0x00095400
-					MHD_OPTION_STRICT_FOR_CLIENT, 1,
-#endif
-#if MHD_VERSION >= 0x00096800
-			MHD_OPTION_SERVER_INSANITY, 1,
-#endif
-					//Enough to send 1MB files through
-					MHD_OPTION_CONNECTION_MEMORY_LIMIT, 1*1024*1024,
-					MHD_OPTION_END);
-#endif
-
-				if (!h2daemon) {
-					wget_error_printf("Cannot start the h2 server.\n");
-					wget_error_printf("HTTP/2 support for MHD not found.\n");
-					return 1;
-				}
-			}
 		}
 #ifdef WITH_GNUTLS_OCSP
 		else {
@@ -831,8 +1306,8 @@ static int _http_server_start(int SERVER_MODE)
 					| MHD_USE_POST_HANDSHAKE_AUTH_SUPPORT
 #endif
 				,
-				port_num, (MHD_AcceptPolicyCallback)_check_to_accept,
-				(void *) (ptrdiff_t) SERVER_MODE, (MHD_AccessHandlerCallback)_answer_to_connection, NULL,
+				port_num, _check_to_accept,
+				(void *) (ptrdiff_t) SERVER_MODE, _answer_to_connection, NULL,
 				MHD_OPTION_HTTPS_CERT_CALLBACK, _ocsp_cert_callback,
 #if MHD_VERSION >= 0x00095400
 				MHD_OPTION_STRICT_FOR_CLIENT, 1,
@@ -881,7 +1356,7 @@ static int _http_server_start(int SERVER_MODE)
 		static char rnd[8] = "realrnd"; // fixed 'random' value
 
 		ocspdaemon = MHD_start_daemon(MHD_USE_SELECT_INTERNALLY,
-			port_num, NULL, NULL, (MHD_AccessHandlerCallback)_ocsp_ahc, NULL,
+			port_num, NULL, NULL, _ocsp_ahc, NULL,
 			MHD_OPTION_DIGEST_AUTH_RANDOM, sizeof(rnd), rnd,
 			MHD_OPTION_NONCE_NC_SIZE, 300,
 #if MHD_VERSION >= 0x00095400
@@ -892,8 +1367,6 @@ static int _http_server_start(int SERVER_MODE)
 #endif
 			MHD_OPTION_CONNECTION_MEMORY_LIMIT, (size_t) 1*1024*1024,
 			MHD_OPTION_END);
-
-		ocsp_resp = wget_malloc(sizeof(struct ocsp_resp_t));
 #endif
 
 		if (!ocspdaemon)
@@ -946,8 +1419,8 @@ static int _http_server_start(int SERVER_MODE)
 		httpsdaemon = MHD_start_daemon(MHD_USE_SELECT_INTERNALLY | MHD_USE_TLS
 				| MHD_USE_POST_HANDSHAKE_AUTH_SUPPORT
 			,
-			port_num, (MHD_AcceptPolicyCallback)_check_to_accept,
-			(void *) (ptrdiff_t) SERVER_MODE, (MHD_AccessHandlerCallback)_answer_to_connection, NULL,
+			port_num, _check_to_accept,
+			(void *) (ptrdiff_t) SERVER_MODE, _answer_to_connection, NULL,
 			MHD_OPTION_HTTPS_CERT_CALLBACK2, _ocsp_stap_cert_callback,
 #if MHD_VERSION >= 0x00095400
 				MHD_OPTION_STRICT_FOR_CLIENT, 1,
@@ -975,10 +1448,6 @@ static int _http_server_start(int SERVER_MODE)
 		else if (SERVER_MODE == OCSP_MODE)
 			dinfo = MHD_get_daemon_info(ocspdaemon, MHD_DAEMON_INFO_BIND_PORT);
 #endif
-#ifdef HAVE_MICROHTTPD_HTTP2_H
-		else if (SERVER_MODE == H2_MODE)
-			dinfo = MHD_get_daemon_info(h2daemon, MHD_DAEMON_INFO_BIND_PORT);
-#endif
 
 		if (!dinfo || dinfo->port == 0)
 			return 1;
@@ -992,7 +1461,7 @@ static int _http_server_start(int SERVER_MODE)
 		else if (SERVER_MODE == OCSP_MODE)
 			ocsp_server_port = port_num;
 #endif
-#ifdef HAVE_MICROHTTPD_HTTP2_H
+#ifdef WITH_LIBNGHTTP2
 		else if (SERVER_MODE == H2_MODE) {
 			h2_server_port = port_num;
 		}
@@ -1012,10 +1481,6 @@ static int _http_server_start(int SERVER_MODE)
 		else if (SERVER_MODE == OCSP_MODE)
 			dinfo = MHD_get_daemon_info(ocspdaemon, MHD_DAEMON_INFO_LISTEN_FD);
 #endif
-#ifdef HAVE_MICROHTTPD_HTTP2_H
-		else if (SERVER_MODE == H2_MODE)
-			dinfo = MHD_get_daemon_info(h2daemon, MHD_DAEMON_INFO_LISTEN_FD);
-#endif
 
 		if (!dinfo)
 			return 1;
@@ -1044,7 +1509,7 @@ static int _http_server_start(int SERVER_MODE)
 					ocsp_server_port = port_num;
 #endif
 
-#ifdef HAVE_MICROHTTPD_HTTP2_H
+#ifdef WITH_LIBNGHTTP2
 				else if (SERVER_MODE == H2_MODE)
 					h2_server_port = port_num;
 #endif
@@ -1121,6 +1586,7 @@ void wget_test_stop_server(void)
 {
 //	wget_vector_free(&response_headers);
 	wget_vector_free(&request_urls);
+	wget_vector_free(&ocsp_responses);
 
 	for (wget_test_url_t *url = urls; url < urls + nurls; url++) {
 		if (url->body_original) {
@@ -1161,7 +1627,7 @@ static char *_insert_ports(const char *src)
 				if (proto_pass == HTTP_1_1_PASS) {
 					dst += wget_snprintf(dst, srclen - (dst - ret), "%d", http_server_port);
 				}
-#ifdef HAVE_MICROHTTPD_HTTP2_H
+#ifdef WITH_LIBNGHTTP2
 				else {
 					dst += wget_snprintf(dst, srclen - (dst - ret), "%d", reject_https_connection ? http_server_port : h2_server_port);
 				}
@@ -1173,7 +1639,7 @@ static char *_insert_ports(const char *src)
 				if (proto_pass == HTTP_1_1_PASS) {
 					dst += wget_snprintf(dst, srclen - (dst - ret), "%d", https_server_port);
 				}
-#ifdef HAVE_MICROHTTPD_HTTP2_H
+#ifdef WITH_LIBNGHTTP2
 				else {
 					dst += wget_snprintf(dst, srclen - (dst - ret), "%d", h2_server_port);
 				}
@@ -1221,8 +1687,8 @@ void wget_test_start_server(int first_key, ...)
 	bool ocsp_stap = 0;
 	bool start_ocsp = 0;
 #endif
-#ifdef HAVE_MICROHTTPD_HTTP2_H
-	bool start_h2 = 1;
+#ifdef WITH_LIBNGHTTP2
+	bool start_h2 = 0;  // Changed to 0: HTTP/2 is now opt-in with WGET_TEST_H2_ONLY
 #endif
 #endif
 
@@ -1243,10 +1709,10 @@ void wget_test_start_server(int first_key, ...)
 #else
 	wget_debug_printf("MHD_OPTION_SERVER_INSANITY: no\n");
 #endif
-#ifdef HAVE_MICROHTTPD_HTTP2_H
-	wget_debug_printf("HAVE_MICROHTTPD_HTTP2_H: yes\n");
+#ifdef WITH_LIBNGHTTP2
+	wget_debug_printf("WITH_LIBNGHTTP2: yes\n");
 #else
-	wget_debug_printf("HAVE_MICROHTTPD_HTTP2_H: no\n");
+	wget_debug_printf("WITH_LIBNGHTTP2: no\n");
 #endif
 #ifdef HAVE_GNUTLS_OCSP_H
 	wget_debug_printf("HAVE_GNUTLS_OCSP_H: yes\n");
@@ -1282,7 +1748,7 @@ void wget_test_start_server(int first_key, ...)
 		case WGET_TEST_HTTP_ONLY:
 #ifdef WITH_TLS
 			start_https = 0;
-#ifdef HAVE_MICROHTTPD_HTTP2_H
+#ifdef WITH_LIBNGHTTP2
 			start_h2 = 0;
 #endif
 #endif
@@ -1291,6 +1757,9 @@ void wget_test_start_server(int first_key, ...)
 			start_http = 0;
 #ifdef WITH_TLS
 			start_https = 0;
+#endif
+#ifdef WITH_LIBNGHTTP2
+			start_h2 = 1;
 #endif
 			break;
 		case WGET_TEST_HTTP_REJECT_CONNECTIONS:
@@ -1325,7 +1794,7 @@ void wget_test_start_server(int first_key, ...)
 			exit(WGET_TEST_EXIT_SKIP);
 #else
 			start_http = 0;
-#ifdef HAVE_MICROHTTPD_HTTP2_H
+#ifdef WITH_LIBNGHTTP2
 			start_h2 = 0;
 #endif
 #ifdef WITH_TLS
@@ -1344,7 +1813,7 @@ void wget_test_start_server(int first_key, ...)
 #ifdef WITH_TLS
 			start_https = 0;
 #endif
-#ifdef HAVE_MICROHTTPD_HTTP2_H
+#ifdef WITH_LIBNGHTTP2
 			start_h2 = 0;
 #endif
 #ifdef WITH_TLS
@@ -1355,7 +1824,7 @@ void wget_test_start_server(int first_key, ...)
 			break;
 #endif
 		case WGET_TEST_SKIP_H2:
-#ifdef HAVE_MICROHTTPD_HTTP2_H
+#ifdef WITH_LIBNGHTTP2
 			start_h2 = 0;
 #endif
 			break;
@@ -1405,11 +1874,11 @@ void wget_test_start_server(int first_key, ...)
 			wget_error_printf_exit("Failed to start HTTPS server, error %d\n", rc);
 	}
 
-#ifdef HAVE_MICROHTTPD_HTTP2_H
-	// start h2 server
+#ifdef WITH_LIBNGHTTP2
+	// start libnghttp2 h2 server
 	if (start_h2) {
-		if ((rc = _http_server_start(H2_MODE)) != 0)
-			wget_error_printf_exit("Failed to start h2 server, error %d\n", rc);
+		if ((rc = h2_server_start()) != 0)
+			wget_error_printf_exit("Failed to start HTTP/2 server, error %d\n", rc);
 	}
 #endif
 #endif
@@ -1496,7 +1965,7 @@ void wget_test_set_executable(const char *program)
 
 void wget_test(int first_key, ...)
 {
-#if !defined WITH_LIBNGHTTP2 || !defined HAVE_MICROHTTPD_HTTP2_H
+#ifndef WITH_LIBNGHTTP2
 	if (!httpdaemon && !httpsdaemon)
 		exit(WGET_TEST_EXIT_SKIP);
 #endif
@@ -1509,8 +1978,12 @@ void wget_test(int first_key, ...)
 #ifndef WITH_LIBNGHTTP2
 			continue;
 #endif
-			if (!h2daemon)
+#ifdef WITH_LIBNGHTTP2
+			if (!nghttp2_server)
 				continue;
+#else
+			continue;  // No HTTP/2 support without libnghttp2
+#endif
 		}
 
 		// now replace {{port}} in the body by the actual server port
@@ -1535,9 +2008,6 @@ void wget_test(int first_key, ...)
 		const char
 			*request_url,
 			*options = "",
-#ifdef WITH_GNUTLS_OCSP
-			*ocsp_resp_file = NULL,
-#endif
 			*executable = global_executable;
 		const wget_test_file_t
 			*expected_files = NULL,
@@ -1575,12 +2045,19 @@ void wget_test(int first_key, ...)
 
 		keep_tmpfiles = 0;
 		clean_directory = 1;
+		h2_min_concurrent_streams = 0;
 
 		if (!request_urls) {
 			request_urls = wget_vector_create(8, NULL);
 			wget_vector_set_destructor(request_urls, NULL);
 		}
 
+#ifdef WITH_GNUTLS_OCSP
+		if (!ocsp_responses) {
+			ocsp_responses = wget_vector_create(2, NULL);
+		}
+#endif
+
 		va_start (args, first_key);
 		for (key = first_key; key; key = va_arg(args, int)) {
 			switch (key) {
@@ -1626,6 +2103,9 @@ void wget_test(int first_key, ...)
 			case WGET_TEST_SERVER_SEND_CONTENT_LENGTH:
 				server_send_content_length = !!va_arg(args, int);
 				break;
+			case WGET_TEST_H2_MIN_CONCURRENT_STREAMS:
+				h2_min_concurrent_streams = va_arg(args, int);
+				break;
 			case WGET_TEST_POST_HANDSHAKE_AUTH:
 				if (va_arg(args, int)) {
 #if MHD_VERSION >= 0x00096302 && GNUTLS_VERSION_NUMBER >= 0x030603
@@ -1633,9 +2113,24 @@ void wget_test(int first_key, ...)
 #endif
 				}
 				break;
-			case WGET_TEST_OCSP_RESP_FILE:
+			case WGET_TEST_OCSP_RESP_FILES:
 #ifdef WITH_GNUTLS_OCSP
-				ocsp_resp_file = va_arg(args, const char *);
+			{
+				const char *ocsp_resp_file = NULL;
+				while ((ocsp_resp_file = va_arg(args, const char *))) {
+					if (ocspdaemon) {
+						ocsp_resp_t ocsp_resp = { .data = NULL, .size = 0 };
+						if (*ocsp_resp_file) {
+							ocsp_resp.data = wget_read_file(ocsp_resp_file, &ocsp_resp.size);
+							if (ocsp_resp.data == NULL) {
+								wget_error_printf_exit("Couldn't read the response from '%s'.\n", ocsp_resp_file);
+							}
+						}
+						wget_vector_add_memdup(ocsp_responses, &ocsp_resp, sizeof(ocsp_resp));
+					}
+				}
+				ocsp_response_pos = 0;
+			}
 #endif
 				break;
 			default:
@@ -1650,19 +2145,6 @@ void wget_test(int first_key, ...)
 			_empty_directory(cmd->data);
 		}
 
-#ifdef WITH_GNUTLS_OCSP
-		if (ocspdaemon) {
-			if (ocsp_resp_file) {
-				ocsp_resp->data = wget_read_file(ocsp_resp_file, &(ocsp_resp->size));
-				if (ocsp_resp->data == NULL) {
-					wget_error_printf_exit("Couldn't read the response.\n");
-				}
-			} else {
-				wget_error_printf_exit("Need value for option WGET_TEST_OCSP_RESP_FILE.\n");
-			}
-		}
-#endif
-
 		// create files
 		if (existing_files) {
 			for (it = 0; existing_files[it].name; it++) {
@@ -1733,7 +2215,7 @@ void wget_test(int first_key, ...)
 					wget_buffer_printf_append(cmd, " \"http://localhost:%d/%s\"",
 					http_server_port, request_url);
 				}
-#ifdef HAVE_MICROHTTPD_HTTP2_H
+#ifdef WITH_LIBNGHTTP2
 				else {
 					wget_buffer_printf_append(cmd, " \"https://localhost:%d/%s\"",
 					h2_server_port, request_url);
@@ -1835,6 +2317,13 @@ void wget_test(int first_key, ...)
 			wget_free(post_handshake_auth);
 #endif
 
+#ifdef WITH_GNUTLS_OCSP
+		for (int i = 0; i < wget_vector_size(ocsp_responses); i++) {
+			ocsp_resp_t *r = wget_vector_get(ocsp_responses, i);
+			wget_xfree(r->data);
+		}
+		wget_vector_clear(ocsp_responses);
+#endif
 		wget_vector_clear(request_urls);
 		wget_buffer_free(&cmd);
 
@@ -1876,10 +2365,10 @@ int wget_test_get_https_server_port(void)
 
 int wget_test_get_h2_server_port(void)
 {
-#ifndef HAVE_MICROHTTPD_HTTP2_H
-	return -1;
-#else
+#ifdef WITH_LIBNGHTTP2
 	return h2_server_port;
+#else
+	return -1;
 #endif
 }
 
