@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2012 Tim Ruehsen
- * Copyright (c) 2015-2022 Free Software Foundation, Inc.
+ * Copyright (c) 2015-2024 Free Software Foundation, Inc.
  *
  * This file is part of Wget.
  *
@@ -40,6 +40,7 @@
 #include <sys/types.h>
 #include <c-ctype.h>
 #include <errno.h>
+#include <inttypes.h>
 
 #include <wget.h>
 #include "../libwget/private.h"
@@ -57,7 +58,7 @@ static void check(int result, int line, const char *msg)
 		ok++;
 	} else {
 		failed++;
-		wget_info_printf("L%d: %s\n", line, msg);
+		wget_error_printf_exit("L%d: %s\n", line, msg);
 	}
 }
 
@@ -273,11 +274,12 @@ static void test_buffer(void)
 	wget_buffer_init(&buf, sbuf, sizeof(sbuf));
 	for (int mid_ws = 0; mid_ws <= 2; mid_ws++) {
 		char expected[16];
-		snprintf(expected, sizeof(expected), "x%.*sy", mid_ws, "  ");
+		snprintf(expected, sizeof(expected), "x%*.*sy", mid_ws, mid_ws, "");
 
 		for (int lead_ws = 0; lead_ws <= 2; lead_ws++) {
 			for (int trail_ws = 0; trail_ws <= 2; trail_ws++) {
-				wget_buffer_printf(&buf, "%.*sx%.*sy%.*s", lead_ws, "  ", mid_ws, "  ", trail_ws, "  ");
+				wget_buffer_printf(&buf, "%*.*sx%*.*sy%*.*s",
+					lead_ws, lead_ws, "", mid_ws, mid_ws, "", trail_ws, trail_ws, "");
 				wget_buffer_trim(&buf);
 				if (!strcmp(buf.data, expected))
 					ok++;
@@ -288,6 +290,18 @@ static void test_buffer(void)
 			}
 		}
 	}
+
+	char expected[] = "";
+	for (int ws = 0; ws <= 3; ws++) {
+		wget_buffer_printf(&buf, "%*.*s", ws, ws, "");
+		wget_buffer_trim(&buf);
+		if (!strcmp(buf.data, expected))
+			ok++;
+		else {
+			failed++;
+			info_printf("test_buffer_trim: got '%s' (expected '%s') (%d)\n", buf.data, expected, ws);
+		}
+	}
 	wget_buffer_deinit(&buf);
 
 	// force reallocation
@@ -302,6 +316,12 @@ static void test_buffer(void)
 	wget_buffer_bufcpy(&buf, bufp);
 	wget_buffer_free(&bufp);
 	wget_buffer_deinit(&buf);
+
+	bufp = wget_buffer_alloc(16);
+	assert(wget_buffer_strcpy(bufp, "moin") == 4);
+	assert(wget_buffer_memset(bufp, 'A', 0) == 0);
+	assert(*bufp->data == 0);
+	wget_buffer_free(&bufp);
 }
 
 static void test_buffer_printf(void)
@@ -521,7 +541,8 @@ static void test_iri_parse(void)
 		const char
 			*userinfo,
 			*password,
-			*host;
+			*host,
+			*safe_uri;
 		uint16_t
 			port;
 		const char
@@ -529,39 +550,48 @@ static void test_iri_parse(void)
 			*query,
 			*fragment;
 	} test_data[] = {
-		{ "1.2.3.4", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "1.2.3.4", 80, NULL, NULL, NULL},
-		{ "1.2.3.4:987", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "1.2.3.4", 987, NULL, NULL, NULL},
-		{ "//example.com/thepath", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", 80, "thepath", NULL, NULL},
+		{ "1.2.3.4", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "1.2.3.4", "http://1.2.3.4", 80, NULL, NULL, NULL},
+		{ "1.2.3.4:987", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "1.2.3.4", "http://1.2.3.4:987", 987, NULL, NULL, NULL},
+		{ "[2a02:2e0:3fe:1001:302::]", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "2a02:2e0:3fe:1001:302::", "http://[2a02:2e0:3fe:1001:302::]", 80, NULL, NULL, NULL},
+		{ "[2a02:2e0:3fe:1001:302::]:987", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "2a02:2e0:3fe:1001:302::", "http://[2a02:2e0:3fe:1001:302::]:987", 987, NULL, NULL, NULL},
+		{ "//example.com/thepath", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", "http:////example.com/thepath", 80, "thepath", NULL, NULL},
 		// { "///thepath", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, NULL, 0, "thepath", NULL, NULL},
-		{ "example.com", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", 80, NULL, NULL, NULL},
-		{ "example.com:555", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", 555, NULL, NULL, NULL},
-		{ "http://example.com", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", 80, NULL, NULL, NULL},
-		{ "http://example.com:", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", 80, NULL, NULL, NULL},
-		{ "http://example.com:/", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", 80, "", NULL, NULL},
-		{ "http://example.com:80/", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", 80, "", NULL, NULL},
-		{ "https://example.com", NULL, WGET_IRI_SCHEME_HTTPS, NULL, NULL, "example.com", 443, NULL, NULL, NULL},
-		{ "https://example.com:443", NULL, WGET_IRI_SCHEME_HTTPS, NULL, NULL, "example.com", 443, NULL, NULL, NULL},
-		{ "https://example.com:444", NULL, WGET_IRI_SCHEME_HTTPS, NULL, NULL, "example.com", 444, NULL, NULL, NULL},
-		{ "http://example.com:80", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", 80, NULL, NULL, NULL},
-		{ "http://example.com:81", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", 81, NULL, NULL, NULL},
-		{ "http://example.com/index.html", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", 80, "index.html", NULL, NULL},
-		{ "http://example.com/index.html?query#frag", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", 80, "index.html", "query", "frag"},
-		{ "http://example.com/index.html?query&param#frag", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", 80, "index.html", "query&param", "frag"},
-		{ "http://example.com/index.html?query&par%26am%61x=1#frag", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", 80, "index.html", "query&par%26am%61x=1", "frag"},
-		{ "http://example.com/index.html?#", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", 80, "index.html", "", ""},
+		{ "example.com", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", "http://example.com", 80, NULL, NULL, NULL},
+		{ "example.com:555", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", "http://example.com:555", 555, NULL, NULL, NULL},
+		{ "http://example.com", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", "http://example.com", 80, NULL, NULL, NULL},
+		{ "http://example.com:", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", "http://example.com:", 80, NULL, NULL, NULL},
+		{ "http://example.com:/", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", "http://example.com:/", 80, "", NULL, NULL},
+		{ "http://example.com:80/", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", "http://example.com:80/", 80, "", NULL, NULL},
+		{ "https://example.com", NULL, WGET_IRI_SCHEME_HTTPS, NULL, NULL, "example.com", "https://example.com", 443, NULL, NULL, NULL},
+		{ "https://example.com:443", NULL, WGET_IRI_SCHEME_HTTPS, NULL, NULL, "example.com", "https://example.com:443", 443, NULL, NULL, NULL},
+		{ "https://example.com:444", NULL, WGET_IRI_SCHEME_HTTPS, NULL, NULL, "example.com", "https://example.com:444", 444, NULL, NULL, NULL},
+		{ "http://example.com:80", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", "http://example.com:80", 80, NULL, NULL, NULL},
+		{ "http://example.com:81", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", "http://example.com:81", 81, NULL, NULL, NULL},
+		{ "http://example.com/index.html", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", "http://example.com/index.html", 80, "index.html", NULL, NULL},
+		{ "http://example.com/index.html?query#frag", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", "http://example.com/index.html?query#frag", 80, "index.html", "query", "frag"},
+		{ "http://example.com/index.html?query&param#frag", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", "http://example.com/index.html?query&param#frag", 80, "index.html", "query&param", "frag"},
+		{ "http://example.com/index.html?query&par%26am%61x=1#frag", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", "http://example.com/index.html?query&par%26am%61x=1#frag", 80, "index.html", "query&par%26am%61x=1", "frag"},
+		{ "http://example.com/index.html?#", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", "http://example.com/index.html?#", 80, "index.html", "", ""},
 #if defined WITH_LIBIDN || defined WITH_LIBIDN2
-		{ "碼標準萬國碼.com", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "xn--9cs565brid46mda086o.com", 80, NULL, NULL, NULL},
+		{ "碼標準萬國碼.com", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "xn--9cs565brid46mda086o.com", "http://碼標準萬國碼.com", 80, NULL, NULL, NULL},
 #endif
 		//		{ "ftp://cnn.example.com&story=breaking_news@10.0.0.1/top_story.htm", NULL,"ftp",NULL,NULL,"cnn.example.com",0,NULL,"story=breaking_news@10.0.0.1/top_story.htm",NULL }
 //		{ "ftp://cnn.example.com?story=breaking_news@10.0.0.1/top_story.htm", NULL, "ftp", NULL, NULL, "cnn.example.com", 0, NULL, "story=breaking_news@10.0.0.1/top_story.htm", NULL},
 //		{ "site;sub:.html", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "site", 0, ";sub:.html", NULL, NULL},
 //		{ "mailto:info@example.com", NULL, "mailto", "info", NULL, "example.com", 0, NULL, NULL, NULL},
-		{ "http://example.com?query#frag", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", 80, NULL, "query", "frag"},
-		{ "http://example.com#frag", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", 80, NULL, NULL, "frag"},
-		{ "http://example.com?#", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", 80, NULL, "", ""},
-		{ "http://example+.com/pa+th?qu+ery#fr+ag", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example+.com", 80, "pa+th", "qu ery", "fr+ag"},
-		{ "http://example.com#frag?x", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", 80, NULL, NULL, "frag?x"},
-		{ "http://user:pw@example.com", NULL, WGET_IRI_SCHEME_HTTP, "user", "pw", "example.com", 80, NULL, NULL, NULL},
+		{ "http://example.com?query#frag", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", "http://example.com?query#frag", 80, NULL, "query", "frag"},
+		{ "http://example.com#frag", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", "http://example.com#frag", 80, NULL, NULL, "frag"},
+		{ "http://example.com?#", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", "http://example.com?#", 80, NULL, "", ""},
+		{ "http://example+.com/pa+th?qu+ery#fr+ag", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example+.com", "http://example+.com/pa+th?qu+ery#fr+ag", 80, "pa+th", "qu ery", "fr+ag"},
+		{ "http://example.com#frag?x", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", "http://example.com#frag?x", 80, NULL, NULL, "frag?x"},
+		{ "http://user:pw@example.com", NULL, WGET_IRI_SCHEME_HTTP, "user", "pw", "example.com", "http://example.com", 80, NULL, NULL, NULL},
+		{ "http://:@example.com", NULL, WGET_IRI_SCHEME_HTTP, "", "", "example.com", "http://example.com", 80, NULL, NULL, NULL},
+		{ "http://user:@example.com", NULL, WGET_IRI_SCHEME_HTTP, "user", "", "example.com", "http://example.com", 80, NULL, NULL, NULL},
+		{ "http://user@example.com", NULL, WGET_IRI_SCHEME_HTTP, "user", NULL, "example.com", "http://example.com", 80, NULL, NULL, NULL},
+		{ "http://:pw@example.com", NULL, WGET_IRI_SCHEME_HTTP, "", "pw", "example.com", "http://example.com", 80, NULL, NULL, NULL},
+		{ "http://user:pw@example@.com", NULL, WGET_IRI_SCHEME_HTTP, "user", "pw", "example@.com", "http://example@.com", 80, NULL, NULL, NULL},
+		{ "http://user:pw@example.com/index.html?query&par%26am%61x=1#frag", NULL, WGET_IRI_SCHEME_HTTP, "user", "pw", "example.com", "http://example.com/index.html?query&par%26am%61x=1#frag", 80, "index.html", "query&par%26am%61x=1", "frag"},
+		{ "http://example.com//path//file", NULL, WGET_IRI_SCHEME_HTTP, NULL, NULL, "example.com", "http://example.com//path//file", 80, "path/file", NULL, NULL},
 	};
 	unsigned it;
 
@@ -574,6 +604,7 @@ static void test_iri_parse(void)
 			|| wget_strcmp(iri->userinfo, t->userinfo)
 			|| wget_strcmp(iri->password, t->password)
 			|| wget_strcmp(iri->host, t->host)
+			|| wget_strcmp(iri->safe_uri, t->safe_uri)
 			|| iri->port != t->port
 			|| wget_strcmp(iri->path, t->path)
 			|| wget_strcmp(iri->query, t->query)
@@ -585,7 +616,9 @@ static void test_iri_parse(void)
 			printf("  display %s (expected %s)\n", iri->display, t->display);
 			printf("  scheme %s (expected %s)\n", wget_iri_scheme_get_name(iri->scheme), wget_iri_scheme_get_name(t->scheme));
 			printf("  user %s (expected %s)\n", iri->userinfo, t->userinfo);
+			printf("  password %s (expected %s)\n", iri->password, t->password);
 			printf("  host %s (expected %s)\n", iri->host, t->host);
+			printf("  safe uri %s (expected %s)\n", iri->safe_uri, t->safe_uri);
 			printf("  port %hu (expected %hu)\n", iri->port, t->port);
 			printf("  path %s (expected %s)\n", iri->path, t->path);
 			printf("  query %s (expected %s)\n", iri->query, t->query);
@@ -1053,8 +1086,8 @@ static void test_parser(void)
 		while ((dp = readdir(dirp)) != NULL) {
 			if (*dp->d_name == '.') continue;
 			if ((ext = strrchr(dp->d_name, '.'))) {
-				char fname[strlen(SRCDIR) + strlen(dp->d_name) + 8];
-				snprintf(fname, sizeof(fname), "%s/files/%s", SRCDIR, dp->d_name);
+				char fname[4096];
+				wget_snprintf(fname, sizeof(fname), "%s/files/%s", SRCDIR, dp->d_name);
 				if (!wget_strcasecmp_ascii(ext, ".xml")) {
 					info_printf("parsing %s\n", fname);
 					wget_xml_parse_file(fname, NULL, NULL, 0);
@@ -1638,7 +1671,7 @@ static void test_utils(void)
 		for (it = 0; it <= 255; it++) {
 			src[0] = (unsigned char) it;
 			wget_memtohex(src, 1, dst1, ndst);
-			snprintf(dst2, ndst, "%02x", src[0]);
+			wget_snprintf(dst2, ndst, "%02x", src[0]);
 			if (strcmp(dst1, dst2)) {
 				info_printf("buffer_to_hex failed: '%s' instead of '%s' (ndst=%d)\n", dst1, dst2, ndst);
 				failed++;
@@ -1960,7 +1993,7 @@ static void test_stringmap(void)
 
 		// now, remove every single entry
 		for (it = 0; it < 26; it++) {
-			snprintf(keybuf, sizeof(keybuf), "http://www.example.com/subdir/%d.html", it);
+			wget_snprintf(keybuf, sizeof(keybuf), "http://www.example.com/subdir/%d.html", it);
 			wget_stringmap_remove(m, keybuf);
 		}
 
@@ -2244,58 +2277,91 @@ static void test_robots(void)
 {
 	static const struct test_data {
 		const char *
-			data;
+			name;
+		const char *
+			input;
 		const char *
 			path[3];
 		const char *
 			sitemap[3];
 	} test_data[] = {
 		{
-			// Deny all robots from part of the server
+			"Allow all + sitemap",
 			"User-agent: *\n"
-			"Disallow: /cgi-bin/\n",
-			{ "/cgi-bin/", NULL },
-			{ "", NULL }
+			"Disallow: # allow all\n"
+			"Sitemap: http://www.example.com/sitemap.xml",
+			{ NULL },
+			{ "http://www.example.com/sitemap.xml", NULL }
 		},
 		{
-			// Deny all robots from the entire server
+			"Allow wget2, bar but deny foo /cgi-bin and deny all /",
+			"User-agent: foo\n"
+			"User-agent: wget2\n"
+			"Disallow: /cgi-bin/\n"
+			"Sitemap: \n"
+			"User-agent: bar\n"
+			"User-agent: wget2\n"
+			"Disallow: \n"
+			"User-agent: *\n"
+			"Disallow: /",
+			{ NULL },
+			{ NULL }
+		},
+		{
+			"Deny all /cgi-bin",
+			"User-agent: *\n"
+			"Disallow: /cgi-bin/ # comment\n",
+			{ "/cgi-bin/", NULL },
+			{ NULL }
+		},
+		{
+			"Deny all /",
 			"User-agent: *\n"
 			"Disallow: /\n",
 			{ "/", NULL },
 			{ NULL }
 		},
 		{
-			// allow a single robot for part of the server
+			"Deny wget2 /cgi/bin",
 			"User-agent: wget2\n"
 			"Disallow: /cgi-bin/\n",
 			{ "/cgi-bin/", NULL },
 			{ NULL }
 		},
 		{
-			// deny a single robot
+			"Deny wget2 /",
 			"User-agent: wget2\n"
 			"Disallow: /\n",
 			{ "/", NULL },
 			{ NULL }
 		},
 		{
-			// allow a single robot
+			"Deny all but allow wget2",
 			"User-agent: *\n"
 			"Disallow: /\n"
 			"User-agent: wget2\n"
 			"Disallow: \n",
-			{ "/", NULL },
+			{ NULL },
 			{ NULL }
 		},
 		{
-			// Allow all robots complete access
-			// or simply don't use a robots.txt at all
+			"Deny all and deny wget2 /cgi-bin/",
+			"User-agent: *\n"
+			"Disallow: /\n"
+			"User-agent: wget2\n"
+			"Disallow: /cgi-bin/",
+			{ "/cgi-bin/", NULL },
+			{ NULL }
+		},
+		{
+			"Allow all",
 			"User-agent: *\n"
 			"Disallow: \n",
 			{ NULL },
 			{ NULL }
 		},
 		{
+			"Deny all /cgi-bin + sitemap",
 			// with 1 sitemap
 			"User-agent: *\n"
 			"Disallow: /cgi-bin/\n"
@@ -2304,7 +2370,23 @@ static void test_robots(void)
 			{ "http://www.example.com/sitemap.xml", NULL }
 		},
 		{
-			// with 2 sitemap
+			"Allow all but deny wget2 /cgi-bin, /tmp + 2 sitemaps",
+			" User-agent : *\n"
+			" Disallow : \n"
+			" User-agent : wget2\n"
+			" User-agent : foo\n"
+			" Disallow : /cgi-bin/#This is a comment\n"
+			" Sitemap : http://www.example1.com/sitemap.xml\n"
+			" Sitemap : \n"
+			" Sitemap : http://www.example2.com/sitemap.xml#Another comment\n"
+			" User-agent : bar\n"
+			" User-agent : wget2\n"
+			" Disallow : /tmp/",
+			{ "/cgi-bin/", "/tmp/", NULL },
+			{ "http://www.example1.com/sitemap.xml", "http://www.example2.com/sitemap.xml", NULL }
+		},
+		{
+			"Deny all /cgi-bin + 2 sitemaps",
 			"User-agent: *\n"
 			"Disallow: /cgi-bin/\n"
 			"Sitemap: http://www.example1.com/sitemap.xml\n"
@@ -2313,7 +2395,7 @@ static void test_robots(void)
 			{ "http://www.example1.com/sitemap.xml", "http://www.example2.com/sitemap.xml", NULL }
 		},
 		{
-			// with 2 "Disallow" entries
+			"Deny all /cgi-bin, /tmp + 2 sitemaps",
 			"User-agent: *\n"
 			"Disallow: /cgi-bin/\n"
 			"Disallow: /tmp/\n"
@@ -2323,7 +2405,7 @@ static void test_robots(void)
 			{ "http://www.example1.com/sitemap.xml", "http://www.example2.com/sitemap.xml", NULL }
 		},
 		{
-			// with equal number entries
+			"Deny all /cgi-bin, /tmp, /jumk + 3 sitemaps",
 			"User-agent: *\n"
 			"Disallow: /cgi-bin/\n"
 			"Disallow: /tmp/\n"
@@ -2335,31 +2417,45 @@ static void test_robots(void)
 			{ "http://www.example1.com/sitemap.xml", "http://www.example2.com/sitemap.xml", "http://www.example3.com/sitemap.xml" }
 		},
 		{
-			// null termination test
+			"Missing EOL",
 			"User-agent: *\n"
 			"Disallow: /cgi-bin/",
 			{ "/cgi-bin/", NULL },
-			{ "", NULL }
+			{ NULL }
 		},
-		{
-			// null termination test #2
-			"User-agent: *"
-			"Disallow: /cgi-bin/",
-			{ "/cgi-bin/", NULL },
-			{ "", NULL }
-		}
 	};
 
 	for (unsigned it = 0; it < countof(test_data); it++) {
 		const struct test_data *t = &test_data[it];
 		wget_robots *robots;
+		int count;
 
-		if (wget_robots_parse(&robots, t->data, PACKAGE_NAME) != WGET_E_SUCCESS) {
-			info_printf("Failed to parse: \"%s\" on robots\n", t->path[it]);
+		if (wget_robots_parse(&robots, t->input, PACKAGE_NAME) != WGET_E_SUCCESS) {
+			info_printf("'%s': Failed to parse input\n", t->name);
 			failed++;
 			continue;
 		}
 
+		count = 0;
+		for (unsigned it2 = 0; it2 < countof(test_data[it].path) && t->path[it2]; it2++, count++)
+			;
+		if (count != wget_robots_get_path_count(robots)) {
+			info_printf("'%s': paths mismatch: expected %d, got %d\n",
+				t->name, count, wget_robots_get_path_count(robots));
+			failed++;
+			goto next;
+		}
+
+		count = 0;
+		for (unsigned it2 = 0; it2 < countof(test_data[it].sitemap) && t->sitemap[it2]; it2++, count++)
+			;
+		if (count != wget_robots_get_sitemap_count(robots)) {
+			info_printf("'%s': sitemap # mismatch: expected %d, got %d\n",
+				t->name, count, wget_robots_get_sitemap_count(robots));
+			failed++;
+			goto next;
+		}
+
 		for (unsigned it2 = 0; it2 < countof(test_data[it].path) && t->path[it2]; it2++) {
 			int n = wget_robots_get_path_count(robots);
 			for (int it3 = 0; it3 < n; it3++) {
@@ -2369,7 +2465,7 @@ static void test_robots(void)
 					it3 = n;
 					ok++;
 				} else if ((strcmp(paths->p, t->path[it2]) && it3 == n - 1)) {
-					info_printf("Cannot find path: \"%s\" on robots\n", t->path[it2]);
+					info_printf("'%s': Cannot find path: \"%s\" on robots\n", t->name, t->path[it2]);
 					failed++;
 				}
 			}
@@ -2384,12 +2480,13 @@ static void test_robots(void)
 					it3 = n;
 					ok++;
 				} else if ((strcmp(sitemaps, t->sitemap[it2]) && it3 == n - 1)) {
-					info_printf("Cannot find sitemap: \"%s\" on robots\n", t->sitemap[it2]);
+					info_printf("'%s': Cannot find sitemap: \"%s\" on robots\n", t->name, t->sitemap[it2]);
 					failed++;
 				}
 			}
 		}
 
+next:
 		wget_robots_free(&robots);
 	}
 }
@@ -2442,6 +2539,82 @@ static void test_set_proxy(void)
 	}
 }
 
+
+static void test_match_no_proxy(void)
+{
+	static const struct test_data {
+		const char *
+			no_proxy;
+		const char *
+			hostip;
+		const char *
+			encoding;
+		int
+			result;
+	} test_data[] = {
+		{ "10.250.192.78/12", "142.251.33.101", NULL, 0},
+		{ "142.250.192.78/12", "142.251.33.101", NULL, 1},
+		{ "142.250.192.78/50", "142.251.33.101", NULL, 0},
+		{ "10.250.192.78.123/12", "142.251.33.101", NULL, 0},
+		{ "142.250.192.78/32", "142.250.180.101", NULL, 0},
+		{ "142.250.192.78/0", "142.250.180.101", NULL, 1},
+		{ "142.250.192.78/-1", "142.250.180.101", NULL, 0},
+		{ "", "142.250.180.101", NULL, 0},
+		{ "142.251.33.101,10.250.192.78/12", "142.251.33.101", NULL, 1},
+		{ "10.250.192.78/12, 142.251.33.101", "142.251.33.101", NULL, 1},
+		{ "10.250.192.78/12, 142.251.33.101", "142.251.33.101", NULL, 1},
+		{ "2402:9400:1234:5670::", "2402:9400:1234:5670::", NULL, 1},
+		{ "2402:9400:1234:5670::", "2402:9400:1234:5671::", NULL, 0},
+		{ "2402:9400:1234:5670::/60", "2402:9400:1234:5678::", NULL, 1},
+		{ "2402:9400:1234:5670::/60", "2402:9400:1234:5680::", NULL, 0}
+	};
+
+	for (unsigned it = 0; it < countof(test_data); it++) {
+		const struct test_data *t = &test_data[it];
+		wget_http_set_no_proxy(t->no_proxy, t->encoding);
+		const wget_vector *no_proxies = wget_http_get_no_proxy();
+		int n = wget_http_match_no_proxy(no_proxies, t->hostip);
+
+		if (n == t->result) {
+			ok++;
+		} else {
+			failed++;
+			info_printf("Failed [%u]: wget_http_match_no_proxy(\"%s\",\"%s\") -> %d (expected %d)\n",
+				it, t->no_proxy, t->hostip, n, t->result);
+		}
+	}
+}
+
+static void test_http_parse_full_date(void) {
+	static const struct test_data {
+		const char *
+				date;
+		uint64_t
+				result;
+	} test_data[] = {
+		{"Sun, 25 May 2003 16:55:12 GMT", 1053881712},
+		{"Wed, 09 Jun 2021 10:18:14 GMT", 1623233894}, // RFC 822 / 1123
+		{"Wednesday, 09-Jun-21 10:18:14", 1623233894}, // RFC 850 / 1036 or Netscape
+		{"Wed, 09-Jun-21 10:18:14", 1623233894}, // RFC 850 / 1036 or Netscape
+		{"Wed Jun 09 10:18:14 2021", 1623233894}, // ANSI C's asctime()
+		{"1 Mar 2027 09:23:12 GMT", 1803892992}, // non-standard
+		{"Sun Nov 26 2023 21:24:47", 1701033887}, // non-standard
+	};
+
+	for (unsigned it = 0; it < countof(test_data); it++) {
+		const struct test_data *t = &test_data[it];
+		uint64_t ts = wget_http_parse_full_date(t->date);
+
+		if (ts == t->result) {
+			ok++;
+		} else {
+			failed++;
+			info_printf("Failed [%u]: wget_http_parse_full_date(\"%s\") -> %"PRIu64" (expected %"PRIu64")\n",
+						it, t->date, ts, t->result);
+		}
+	}
+}
+
 // Add some corner cases here.
 static void test_parse_header_line(void)
 {
@@ -2542,14 +2715,14 @@ int main(int argc, const char **argv)
 		// fallthrough
 	}
 	else if (!strcmp(valgrind, "1")) {
-		char cmd[strlen(argv[0]) + 256];
+		char cmd[4096];
 
-		snprintf(cmd, sizeof(cmd), "VALGRIND_TESTS=\"\" valgrind --error-exitcode=301 --leak-check=yes --show-reachable=yes --track-origins=yes %s", argv[0]);
+		wget_snprintf(cmd, sizeof(cmd), "VALGRIND_TESTS=\"\" valgrind --error-exitcode=301 --leak-check=yes --show-reachable=yes --track-origins=yes %s", argv[0]);
 		return system(cmd) != 0;
 	} else {
-		char cmd[strlen(valgrind) + strlen(argv[0]) + 32];
+		char cmd[4096];
 
-		snprintf(cmd, sizeof(cmd), "VALGRIND_TESTS="" %s %s", valgrind, argv[0]);
+		wget_snprintf(cmd, sizeof(cmd), "VALGRIND_TESTS="" %s %s", valgrind, argv[0]);
 		return system(cmd) != 0;
 	}
 
@@ -2587,7 +2760,7 @@ int main(int argc, const char **argv)
 	test_iri_relative_to_absolute();
 	test_iri_compare();
 	test_parser();
-
+	test_match_no_proxy();
 	test_cookies();
 	test_hsts();
 	test_hpkp();
@@ -2598,6 +2771,7 @@ int main(int argc, const char **argv)
 	test_set_proxy();
 	test_parse_response_header();
 	test_parse_header_line();
+	test_http_parse_full_date();
 
 	selftest_options() ? failed++ : ok++;
 
